/* ####################################################
   # Copyright 2002, 2003 Ren Serral,                #
   #                                                  #
   # correlation.c - Correlation module for Netmeter. #
   #       v2.0 (04-2002)                             #
   #       v2.01(02-2003)                             #
   ####################################################
   
   This file is part of NetMeter.

   NetMeter is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   NetMeter is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with NetMeter; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>

#include <math.h>

#include "lib/common.h"

#define CHUNK 5000

/* Procedure for displaying usage information */
void usage();

/* Global Variable declaration */
int sh_max, sh_min, sh_avg, sh_values;
int is_verbose;

int main(int argc, char **argv) {
	int sel_opt;
	int fd_read1, fd_read2, fd_write; // File Handlers
	int ok, i;
	long int n_elements; // Holds the number of columns of the input data
	double first_value, second_value; // Double input
	double *d_array[2], *d_tmp[2]; // Two column array

	double corr_xy, corr_xx, corr_yy;
	double x_sum, y_sum, x2_sum, xy_sum, A, correlation;
	double avg_x, avg_y, min_v1, min_v2, max_v1, max_v2;
	
	int has_input = 0, has_output = 0, has_force = 0; // Boolean switches
	char *out_file; // Output file
	
	char c_tmp[200];

	// Inicialitzations
	fd_read1  = STDIN_FILENO;
	fd_read2  = STDIN_FILENO;
	fd_write  = STDOUT_FILENO;

	d_array[0] = malloc(CHUNK * sizeof(double));
	d_array[1] = malloc(CHUNK * sizeof(double));
	bzero(d_array[0], CHUNK * sizeof(double));
	bzero(d_array[1], CHUNK * sizeof(double));

	n_elements = 0;
	avg_x = 0; avg_y = 0;
	x_sum = 0; y_sum = 0; xy_sum = 0; x2_sum = 0;
	A = 0; correlation = 0;
	corr_xy = 0; corr_xy = 0; corr_xx = 0; corr_yy = 0;
	
	sh_max = 0; sh_min = 0; sh_avg = 0; sh_values = 0;
	is_verbose = 0;

	// Main option parsing loop
	while(1) {
		sel_opt = getopt(argc, argv, "i:o:vfhmMaV");
		if (sel_opt == -1)
			break;
		switch (sel_opt) {
		case 'i':
			if (has_input >= 1) {
				fd_read2 = open(optarg,O_RDONLY);
				if (fd_read1 < 0) {
					perror("Error opening input file 2");
					exit(-1);
				}
				has_input++;
				break;
			}
			else if (has_input >= 2) {
				fprintf(stderr, "Ignoring third input file");
				break;
			}
			fd_read1 = open(optarg,O_RDONLY);
			fd_read2 = fd_read1;
			if (fd_read1 < 0) {
				perror("Error opening input file 1");
				exit(-1);
			}
			has_input = 1;
		break;
		case 'o':
			if (has_output)
				break;
			out_file = malloc(strlen(optarg+1));
			strncpy(out_file,optarg,strlen(optarg));
			has_output = 1;
		break;
		case 'v':
			sh_max = 1; sh_min = 1; sh_avg = 1; sh_values = 1;
			is_verbose = 1;
		break;
		case 'a':
			sh_avg = 1;
		break;
		case 'M':
			sh_max = 1;
		break;
		case 'm':
			sh_min = 1;
		break;
		case 'V':
			sh_values = 1;
		break;
		case 'f':
			has_force = 1;
		break;
		case 'h':
			usage();
			if (fd_read1 != STDIN_FILENO)
				close(fd_read1);
			if (fd_read2 != fd_read1)
				close(fd_read2);
			exit(0);
		break;
		default:
			usage();
			exit(-2);
		break;
		} // End Switch sel_opt
	} // End While (option parsing)
	if (has_output) {
		if (!has_force) {
			fd_write = open(out_file, O_RDONLY);
			if (fd_write > 0) {
				close(fd_write);
				close(fd_read1);
				if (fd_read2 != fd_read1)
					close(fd_read2);
				fprintf(stderr,"Error: Output file %s exists, use -f to override\n", out_file);
				exit(-1);
			}
		}
		fd_write = open(out_file, O_RDWR | O_CREAT, 0666);
		if (fd_write < 0) {
			perror("Error creating output file");
			exit(-1);
		}
	} // End if has_output

	do {
		ok = read_double(fd_read1, &first_value, is_verbose);
		if (ok >= 0) {
			/* Ask for memory only if is really needed
			   ==> if we don't get here execution'll be aborted
			 */
			if (!(n_elements % CHUNK) && n_elements != 0) {
				d_tmp[0] = malloc((n_elements + CHUNK) * sizeof(double));
				d_tmp[1] = malloc((n_elements + CHUNK) * sizeof(double));
				memcpy(d_tmp[0],d_array[0],n_elements * sizeof(double));
				memcpy(d_tmp[1],d_array[1],n_elements * sizeof(double));
				bzero((d_tmp[0]+n_elements), CHUNK * sizeof(double));
				bzero((d_tmp[1]+n_elements), CHUNK * sizeof(double));
				free(d_array[1]);
				free(d_array[0]);
				d_array[0] = d_tmp[0];
				d_array[1] = d_tmp[1];
				d_tmp[0] = NULL;
				d_tmp[1] = NULL;
			}
			d_array[0][n_elements] = first_value;
		}
		else
			break;
		ok = read_double(fd_read2, &second_value, is_verbose);
		if (ok >= 0)
			d_array[1][n_elements] = second_value;
		else
			break;
		if (n_elements == 0) {
			max_v1 = d_array[0][n_elements];
			max_v2 = d_array[1][n_elements];
			min_v1 = d_array[0][n_elements];
			min_v2 = d_array[1][n_elements];
		}
		else {
			if (d_array[0][n_elements] > max_v1)
				max_v1 = d_array[0][n_elements];
			if (d_array[1][n_elements] > max_v2)
				max_v2 = d_array[1][n_elements];
			if (d_array[0][n_elements] < min_v1)
				min_v1 = d_array[0][n_elements];
			if (d_array[1][n_elements] < min_v2)
				min_v2 = d_array[1][n_elements];
		}

		// To avoid another loop let's accumulate some values
		x_sum  += d_array[0][n_elements];	// x accumulator
		y_sum  += d_array[1][n_elements];	// y accumulator
		xy_sum += d_array[0][n_elements] * d_array[1][n_elements];	// x*y accumulator
		x2_sum += d_array[0][n_elements] * d_array[0][n_elements];	// x accumulator
		n_elements++; // Number of elements in the array
	} while(ok > 0);
	// -- Here starts the real correlation calculations
	A = (double)(-(x_sum * y_sum) + (n_elements * xy_sum))/(double)((n_elements * x2_sum) - (x_sum * x_sum));
	avg_x = x_sum / n_elements;
	avg_y = y_sum / n_elements;
	
	for (i = 0; i < n_elements; i++) {
		corr_xy += (d_array[0][i] - avg_x) * (d_array[1][i] - avg_y); // WATCH!!!! Not final value
		corr_xx += (d_array[0][i] - avg_x) * (d_array[0][i] - avg_x);
		corr_yy += (d_array[1][i] - avg_y) * (d_array[1][i] - avg_y);
	}

	corr_xy = corr_xy / n_elements;
	corr_xy = corr_xy * corr_xy;
	corr_xx = corr_xx / n_elements;
	corr_yy = corr_yy / n_elements;

	correlation = sqrt(corr_xy / (corr_xx * corr_yy));
	if(A < 0)
		correlation = -correlation;

	if (fd_write > STDOUT_FILENO) {
		if (sh_values) {
			sprintf(c_tmp,"\tTest done with %d values\n",n_elements);
			write(fd_write,c_tmp,strlen(c_tmp));
		}
		if (sh_max) {
			sprintf(c_tmp,"\tMax. Value test 1: %f; Max. Value test 2: %f\n",max_v1,max_v2);
			write(fd_write,c_tmp,strlen(c_tmp));
		}
		if (sh_min) {
			sprintf(c_tmp,"\tMin. Value test 1: %f; Min. Value test 2: %f\n",min_v1,min_v2);
			write(fd_write,c_tmp,strlen(c_tmp));
		}
		if (sh_avg) {
			sprintf(c_tmp,"\tAverage Value test 1: %f; Average Value test 2: %f\n",avg_x, avg_y);
			write(fd_write,c_tmp,strlen(c_tmp));
		}
		if (sh_values || sh_max || sh_min || sh_avg) {
			sprintf(c_tmp, "The Correlation value is: %2.3f\n", correlation);
			write(fd_write,c_tmp,strlen(c_tmp));
		}
		else {
			sprintf(c_tmp, "%2.3f\n", correlation);
			write(fd_write,c_tmp,strlen(c_tmp));
		}
	}
	else {
		if (sh_values) {
			fprintf(stdout,"\tTest done with %d values\n",n_elements);
		}
		if (sh_max) {
			fprintf(stdout,"\tMax. Value test 1: %f; Max. Value test 2: %f\n",max_v1,max_v2);
		}
		if (sh_min) {
			fprintf(stdout,"\tMin. Value test 1: %f; Min. Value test 2: %f\n",min_v1,min_v2);
		}
		if (sh_avg) {
			fprintf(stdout,"\tAverage Value test 1: %f; Average Value test 2: %f\n",avg_x, avg_y);
		}
		if (sh_values || sh_max || sh_min || sh_avg) {
			fprintf(stdout, "The Correlation value is: %2.3f\n", correlation);
		}
		else {
			fprintf(stdout, "%2.3f\n", correlation);
		}
	}
	free(d_array[1]);
	free(d_array[0]);
	close(fd_read1);
	close(fd_read2);
	close(fd_write);
	return 0;
}

/* Procedure for displaying usage information */
void usage() {
	printf("Usage: ./correlation -i <input_file> -o <output_file> [-f] [-v]\n");
	printf("	-i   input_file, if is omitted stdin will be used\n");
	printf("	-o   output_file, if is omitted stdout will be used\n");
	printf("	-f   for overwriting destination file\n");
	printf("	-V   Show the number of values of each variable\n");
	printf("	-a   Show data average for both variables\n");
	printf("	-m   Show minimum for both variables\n");
	printf("	-M   Show maximum for both variables\n");
	printf("	-v   Be verbose (implies -a -V -m -M\n");
}

