/* ####################################################
   # Copyright 2002, 2003 Ren Serral,                #
   #                                                  #
   # common.c - Library for common functions.         #
   #       v2.0 (04-2002)                             #
   #       v2.01(02-2003)                             #
   ####################################################

   This file is part of NetMeter.

   NetMeter is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   NetMeter is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with NetMeter; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>

#include "common.h"

/* Function to read a double from the input stream */
int read_double(int fd, double *fl, int is_verbose) {
	int tmp, c;
	int in_decimal = 0, is_neg = 0;

	if (!fl) {
		if (is_verbose)
			fprintf(stderr, "read_double: Error in parameters, fl can't be NULL\n");
		exit(-MYEINVAL);
	}
	if (fd < 0) {
		if (is_verbose)
			fprintf(stderr, "read_double: Error in file descriptor\n");
		exit(-ENOFD);
	}
	*fl = 0;
	do {
		tmp = read(fd,&c,1);
	} while (is_separator((char)c) && tmp);  // Eat all separators
	if (!tmp)
		return -EEOF;
	if ((char)c == '-') {
		is_neg = 1;
		tmp = read(fd,&c,1);
	}
	do {
		if (is_number((char)c)) {
			if (in_decimal > 0) {
				*fl += to_integer((char)c) / pow(10,in_decimal);
				in_decimal++;
			}
			else {
				*fl = *fl * 10 + to_integer((char)c);
			}
		}
		else if ((char)c == ',' || (char)c == '.') {
			in_decimal = 1;
		}
		else {
			if (is_neg)
				*fl = *fl * -1;
			if (is_separator((char)c) && tmp)
				return (char)c;
			return 0;
		}
		tmp = read(fd,&c,1);
	} while (tmp == 1);
	if (is_neg)
		*fl = *fl * -1;
	return 0;
}

/* Function to read a line of text from the input stream */
int read_line(int fd, char *fl, int is_verbose) {
	int tmp, c, i;

	if (!fl) {
		if (is_verbose)
			fprintf(stderr, "read_line: Error in parameters, fl can't be NULL\n");
		exit(-MYEINVAL);
	}
	if (fd < 0) {
		if (is_verbose)
			fprintf(stderr, "read_line: Error in file descriptor\n");
		exit(-ENOFD);
	}
	*fl = 0;
	do {
		tmp = read(fd,&c,1);
	} while (is_separator((char)c) && tmp);  // Eat all separators
	if (!tmp)
		return -EEOF;
	i = 0;
	do {
		fl[i] = c;
		i++;
		if (i >= MAX_LINE)
			return -EBIGLINE;
		tmp = read(fd,&c,1);
	} while (tmp == 1 && (char )c != '\n');
	fl[i]='\0';
	return i;
}

/* Function to write a line of text to the ouput stream fd*/
int write_line(char *fl, int fd) {
	int i = 0;

	if (!fl)
		return -MYEINVAL;
	while (fl[i] != '\n' && fl[i] != '\0')
		i++;
	return write(fd, fl, i);
}

int to_integer(char c) {
	if (c < '0' || c > '9')
		return -ENOINT;
	return c - '0';
}

int to_string(long long number, unsigned int size, char *result) {
	char tmp[MAX_LINE];
	int sizeNow = 0, module, totalSize;
	long long current = number;

	if (!result)
		return -MYEINVAL;
	do {
		module = current % 10;
		tmp[sizeNow] = (char)module + '0';
		sizeNow++;
	} while ((current = current / 10) != 0 && sizeNow < size);
	totalSize = sizeNow;

	for(current = size - totalSize; current > 0; current--)
		result[current-1]='0';
	while (sizeNow) {
		result[(size - totalSize) + sizeNow - 1] = tmp[totalSize - sizeNow];
		sizeNow--;
	}
	if ((totalSize + 1) < size)
		result[totalSize + 1]='\0';
}

int is_separator(char c) {
	char c2[2];

	c2[0] = c;
	c2[1] = '\0';

	if (!strstr(C_SEPARATOR, c2)) {
		return 0;
	}
	return 1;
}

int is_number(char c) {
	if (c >= '0' && c <= '9')
		return 1;
	return 0;
}

int is_hex(char c) {
	if ((c >= '0' && c <= '9') || (c >= 'a' && c <= 'f') || (c >= 'A' && c <= 'F'))
		return 1;
	return 0;
}

int getInteger(char *data) {
	int number, i;
	if (!data) {
#ifdef DEBUG
		fprintf(stderr, "getInteger(char *data): Invalid data\n");
#endif
		return -MYEINVAL;
	}
	number = 0;
	i = 0;
	while (data[i] && is_number(data[i])) {
		number = number * 10 + to_integer(data[i]);
		i++;
	}
	return number;
}

int getIP(char *data, p_address addr) {
	int i, j, type;
	char *tmp = data;
	char ip[40];
	if (!data) {
#ifdef DEBUG
		fprintf(stderr, "getIP(char *data, address addr): Invalid data\n");
#endif
		return -MYEINVAL;
	}
	if (!addr) {
#ifdef DEBUG
		fprintf(stderr, "getIP(char *data, address addr): Invalid pointer\n");
#endif
		return -MYEINVAL;
	}
	i = 0;
	j = 0;
	while (is_separator(tmp[i]))
		i++;
	while (tmp[i] && is_number(tmp[i])) {
		ip[j] = tmp[i];
		i++;
		j++;
	}
	if (tmp[i] == '.') { // IPv4 Mode
		type = IPv4;
		if (getInteger(ip) > 255) {
			return -MYEINVAL;
		}
		ip[j] = '.';
		j++;
		i++;
		while (tmp[i] && is_number(tmp[i])) {
			ip[j] = tmp[i];
			i++;
			j++;
		}
		if (getInteger(ip) > 255) {
			return -MYEINVAL;
		}
		ip[j] = '.';
		j++;
		i++;
		while (tmp[i] && is_number(tmp[i])) {
			ip[j] = tmp[i];
			i++;
			j++;
		}
		if (getInteger(ip) > 255) {
			return -MYEINVAL;
		}
		ip[j] = '.';
		j++;
		i++;
		while (tmp[i] && is_number(tmp[i])) {
			ip[j] = tmp[i];
			i++;
			j++;
		}
		ip[j] = '\0';
		inet_aton(ip, &(addr->s_address.sin_addr));
	}
	else {		// IPv6 Mode
		int error;
		struct addrinfo hints;
		struct addrinfo *res;
		type = IPv6;
		while (is_hex(tmp[i]) || tmp[i] == ':') {
			ip[j] = tmp[i];
			j++;
			i++;
		}
		ip[j] = '\0';
		// IPv6 doesn't have any inet_aton, so that's a workaround
		memset(&hints, 0, sizeof(hints));
		/* set-up hints structure */
		hints.ai_family = PF_INET6;
		hints.ai_flags = AI_NUMERICHOST;
		error = getaddrinfo(ip, NULL, &hints, &res);
		if (error) {
			perror(gai_strerror(error));
			freeaddrinfo(res);
			return;
		}
		else {
			memcpy(&(addr->s6_address), (struct sockaddr_in6 *)res->ai_addr,
				sizeof(struct sockaddr_in6));
		}
		freeaddrinfo(res);
	}
	return type;
}

int getTime(char *data, struct timeval *timestamp) {
	char *tmp = data;
	char value[8];
	int i, j;
	if (!data) {
#ifdef DEBUG
		fprintf(stderr, "getTime(char *data, timeval *timestamp): Invalid data\n");
#endif
		return -MYEINVAL;
	}
	i = 0;
	j = 0;
	while (is_separator(tmp[i]))
		i++;
	while (tmp[i] && is_number(tmp[i])) {
		value[j] = tmp[i];
		i++;
		j++;
	}
	if (tmp[i] != ':') {
		return -MYEINVAL;
	}
	i++;
	timestamp->tv_sec = getInteger(value) * 60 * 60;
	bzero(value, 8);
	j = 0;
	while (tmp[i] && is_number(tmp[i])) {
		value[j] = tmp[i];
		i++;
		j++;
	}
	if (tmp[i] != ':') {
		return -MYEINVAL;
	}
	i++;
	timestamp->tv_sec += getInteger(value) * 60;
	bzero(value, 8);
	j = 0;
	while (tmp[i] && is_number(tmp[i])) {
		value[j] = tmp[i];
		i++;
		j++;
	}
	if (tmp[i] != '.') {
		return -MYEINVAL;
	}
	i++;
	timestamp->tv_sec += getInteger(value);
	timestamp->tv_usec = getInteger(&(tmp[i]));
	return OK;
}

int time_diff(struct timeval *time1, struct timeval *time2, struct timeval *result) {
	long int li_tmp;
	struct timeval ts_fake;
	if (!time1 || !time2 || !result) {
#ifdef DEBUG
		fprintf(stderr, "int time_diff(timeval *time1, timeval *time2, struct timeval *result): Invalid NULL pointer\n");
#endif
		return -MYEINVAL;
	}
	if (time1->tv_sec - time2->tv_sec < 0) {
#ifdef DEBUG
		fprintf(stderr, "Synchronization error!!!!!!!");
#endif
		return -MYEINVAL;
	}
	
	li_tmp = (time1->tv_sec * 1000000 + time1->tv_usec) - (time2->tv_sec * 1000000 + time2->tv_usec);
	ts_fake.tv_sec = li_tmp / 1000000;
	ts_fake.tv_usec = li_tmp % 1000000;
	memcpy(result, &ts_fake, sizeof(struct timeval));
	return OK;
}

int time_sum(struct timeval *time1, struct timeval *time2, struct timeval *result) {
	struct timeval ts_fake;
	if (!time1 || !time2 || !result) {
#ifdef DEBUG
		fprintf(stderr, "int time_sum(timeval *time1, timeval *time2, struct timeval *result): Invalid NULL pointer\n");
#endif
		return -MYEINVAL;
	}
	ts_fake.tv_usec = time1->tv_usec + time2->tv_usec;
	if (ts_fake.tv_usec >= 1000000) {
		ts_fake.tv_sec = time1->tv_sec + time2->tv_sec + (ts_fake.tv_usec / 1000000);
		ts_fake.tv_usec = ts_fake.tv_usec % 1000000;
	} else
		ts_fake.tv_sec = time1->tv_sec + time2->tv_sec;
	memcpy(result, &ts_fake, sizeof(struct timeval));
	return OK;
}

int time_div(struct timeval *time1, long int divisor, struct timeval *result) {
	long int li_tmp;
	struct timeval ts_fake;
	if (!time1 || !result) {
#ifdef DEBUG
		fprintf(stderr, "int time_div(timeval *time1, int divisor, struct timeval *result): Invalid NULL pointer\n");
#endif
		return -MYEINVAL;
	}
	if (!divisor) {
#ifdef DEBUG
		fprintf(stderr, "int time_div(timeval *time1, int divisor, struct timeval *result): Can't divide by 0\n");
#endif
		return -MYEINVAL;
	}
	li_tmp = (time1->tv_sec * 1000000 + time1->tv_usec) / divisor;
	ts_fake.tv_sec = li_tmp / 1000000;
	ts_fake.tv_usec = li_tmp % 1000000;
	memcpy(result, &ts_fake, sizeof(struct timeval));
	return OK;
}

int ts_is_bigger(struct timeval *time1, struct timeval *time2) {
	if (!time1 || !time2) {
#ifdef DEBUG
		fprintf(stderr, "int tx_is_bigger(struct timeval *time1, struct timeval *time2): Invalid NULL pointer\n");
#endif
		return -MYEINVAL;
	}
	if (time1->tv_sec > time2->tv_sec)
		return TRUE;
	else if (time1->tv_sec < time2->tv_sec)
		return FALSE;
	else if (time1->tv_usec > time2->tv_usec)
		return TRUE;
	else
		return FALSE;
}

int ts_is_bigger_or_equal(struct timeval *time1, struct timeval *time2) {
	if (!time1 || !time2) {
#ifdef DEBUG
		fprintf(stderr, "int tx_is_bigger(struct timeval *time1, struct timeval *time2): Invalid NULL pointer\n");
#endif
		return -MYEINVAL;
	}
	if (time1->tv_sec >= time2->tv_sec)
		return TRUE;
	else if (time1->tv_sec < time2->tv_sec)
		return FALSE;
	else if (time1->tv_usec >= time2->tv_usec)
		return TRUE;
	else
		return FALSE;
}

int ts_is_smaller(struct timeval *time1, struct timeval *time2) {
	if (!time1 || !time2) {
#ifdef DEBUG
		fprintf(stderr, "int tx_is_smaller(struct timeval *time1, struct timeval *time2): Invalid NULL pointer\n");
#endif
		return -MYEINVAL;
	}
	if (time1->tv_sec > time2->tv_sec)
		return FALSE;
	else if (time1->tv_sec < time2->tv_sec)
		return TRUE;
	else if (time1->tv_usec >= time2->tv_usec)
		return FALSE;
	else
		return TRUE;
}

int ts_is_smaller_or_equal(struct timeval *time1, struct timeval *time2) {
	if (!time1 || !time2) {
#ifdef DEBUG
		fprintf(stderr, "int tx_is_smaller(struct timeval *time1, struct timeval *time2): Invalid NULL pointer\n");
#endif
		return -MYEINVAL;
	}
	if (time1->tv_sec > time2->tv_sec)
		return FALSE;
	else if (time1->tv_sec <= time2->tv_sec)
		return TRUE;
	else if (time1->tv_usec > time2->tv_usec)
		return FALSE;
	else
		return TRUE;
}

int ts2double(struct timeval *time1, double *result) {
	if (!time1 || !result)
		return -MYEINVAL;
	*result = time1->tv_sec * 1000 + time1->tv_usec / 1000;
	return OK;
}

int ts2char(struct timeval *time1, char result[TSSIZE]) {

	if (!time1 || !result)
		return -MYEINVAL;
	bzero(result, TSSIZE);
	to_string(time1->tv_sec / 3600, 2, &result[0]);
	result[2] = ':';
	to_string(time1->tv_sec % 3600 / 60, 2, &result[3]);
	result[5] = ':';
	to_string((time1->tv_sec % 3600) % 60, 2, &result[6]);
	result[8] = '.';
	to_string(time1->tv_usec, 6, &result[9]);
	result[TSSIZE-1] = '\0';
	return OK;
}

int to_msec(struct timeval *time1, double *result) {
	if (!time1 || !result) {
#ifdef DEBUG
		fprintf(stderr, "int to_msec(struct timeval *time1, double *result): Invalid NULL pointer\n");
#endif
		return -MYEINVAL;
	}
	*result = time1->tv_sec*1000 + (time1->tv_usec/1000);
	return OK;
}

char *inet6_ntoa(struct sockaddr_in6 *ss) {
	int error;
	char *name;
	char namebuf[MAX_LINE];

	error = getnameinfo((struct sockaddr *)ss, sizeof(struct sockaddr_in6), namebuf, sizeof(namebuf), NULL, 0, NI_NUMERICHOST);
	if (error)
		perror("getnameinfo");
	name = namebuf;
}
