/* ####################################################
   # Copyright 2002, 2003 Ren Serral,                #
   #                                                  #
   # mos.c - VoIP MOS calculation as                  #
   #         stated onsaid on ITU-T G.107             #
   #       v2.0 (04-2002)                             #
   #       v2.01(02-2003)                             #
   ####################################################

   This file is part of NetMeter.

   NetMeter is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   NetMeter is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with NetMeter; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <time.h>

#include <math.h>

#include "lib/common.h"
#include "lib/flows.h"

#include "mos.h"

#define CHUNK 5000

/* Temporary data */
double No(struct parameters *params) {
	double Nr1, Nr2, Pro, Pr1, Nfo;
	int LSTR;
	if (!params)
		return 0;
	Nr1 = params->Ps - params->SLR - params->Ds - 100 + 0.004 *
			pow(params->Ps - (params->SLR + params->RLR) - params->Ds - 14, 2);
	LSTR = params->STMR + params->Dr;
	Pro = params->Pr + 10 * log(1 + pow(10, (10 - LSTR) / 10));
	Pr1 = Pro + 0.008 * pow(Pro - 35, 2);
	Nr2 = Pr1 - 121 + params->RLR;
	Nfo = params->Nfor + params->RLR;
	return 10 * log(pow(10, Nr1 / 10) + pow(10, Nr2 / 10) +
			pow(10, params->Nc / 10) + pow(10, Nfo / 10)) / log(10);
}

/* Formula 3 to 7 of the ITU-T G.107 Specs
 */
double Nt(struct parameters *params) {
	return No(params) - params->RLR;
}

/* Formula 2 of the ITU-T G.107 Specs
 */
double Ro(struct parameters *params) {
	if (!params)
		return 0;
	return 15 - 1.5 * (params->SLR + No(params));
}

/* Formula 9 and 10 of the ITU-T G.107 Specs
 */
double Iolr(struct parameters *params) {
	double Xolr;
	if (!params)
		return 0;
	Xolr = params->SLR + params->RLR + 0.2 * (64 + Nt(params));
	return 20 * (pow(pow(1 + (Xolr / 8), 8), 1 / 8) - Xolr / 8);
}

/* Formula 11 to 12 of the ITU-T G.107 Specs
 */
double Ist(struct parameters *params) {
	double STMRo;
	double tmp;
	if (!params)
		return 0;
	STMRo = -10 * log(pow(10, params->STMR / 10) + pow(10, params->TELR / 10) * exp(params->T / 4));
	tmp = 12 * (1 + pow(pow((STMRo - 13) / 6, 8), 1 / 8));
	tmp = tmp - 28 * (1 + pow(pow((STMRo + 1) / 19.4, 35), 1 / 35));
	return tmp - 13 * (1 + pow(pow((STMRo - 3) / 33, 13), 1 / 13)) + 29;
}

/* Formula 13 to 17 of the ITU-T G.107 Specs
 */
double Iq(struct parameters *params) {
	double Q;
	double G;
	if (!params)
		return 0;
	if (params->qdu < 1)
		params->qdu = 1;
	Q = 37 - 15 * log(params->qdu);
	G = 1.07 + 0.258 * Q + 0.0602 * pow(Q, 2);
	return 15 * log(1 + pow(10, (Ro(params) - 100) / 15 + 46 / 8.4 - G / 9) +
			pow(10, 46 / 30 - G / 40));
}

/* Formula 8 of the ITU-T G.107 Specs
 */
double Is(struct parameters *params) {
	if (!params)
		return 0;
	return Iolr(params) + Ist(params) + Iq(params);
}

/* Formula 22 and 23 of the ITU-T G.107 Specs
 */
double TERV(struct parameters *params) {
	double tmp;
	if (!params)
		return 0;
	tmp = params->TELR + 6 * exp(-0.3 * pow(params->T, 2)) - 40 *
			log((1 + params->T / 10) / (1 + params->T / 150));
	if (params->STMR < 9)
		return tmp + 0.5 * Ist(params);
	return tmp;
}

/* Formula 19, 21 and 24 of the ITU-T G.107 Specs
 */
double Idte(struct parameters *params) {
	double Re, Roe, Xdt, tmp;
	if (!params)
		return 0;
	Re = 80 + 2.5 * (TERV(params) - 14);
	Roe = -1.5 * (No(params) - params->RLR);
	Xdt = (Roe - Re) / 2;
	tmp = Xdt + sqrt(pow(Xdt, 2) + 100);
	tmp = (tmp - 1) * (1 - exp(-params->T));
	if (params->STMR > 20)
		return sqrt(pow(tmp, 2) + pow(Ist(params), 2));
	return tmp;
}

/* Formula 25 and 26 of the ITU-T G.107 Specs
 */
double Idle(struct parameters *params) {
	double Rle, Xdl;
	if (!params)
		return 0;
	Rle = 10.5 * (params->WEPL + 7) * pow(params->Tr + 1, -1/4);
	Xdl = (Ro(params) - Rle) / 2;
	return Xdl + sqrt(pow(Xdl, 2) + 169);
}

/* Formula 27 and 28 of the ITU-T G.107 Specs
 */
double Idd(struct parameters *params) {
	double X;
	if (!params)
		return 0.0;
	if (params->Ta <= 100)
		return 0.0;
	else {
		X = log(params->Ta / 100) / log(2);
		return 25 * (pow(1 + pow(X, 6), 1/6) - 3 * pow(1 + pow(X / 3, 6), 1 / 6) + 2);
	}
}

/* Formula Id ITU-T G.107 Specs
 */
double Id(struct parameters *params) {
	if (!params)
		return 0;
	return Idte(params) + Idle(params) + Idd(params);
}

/* Formula 29, packet loss inclusion of the ITU-T G.107 Specs
 */
double Ieef(struct parameters *params) {
	if (!params)
		return 0;
	if (params->Ppl)
		return params->Ie + (95 - params->Ie) * (params->Ppl / (params->Ppl / params->BurstR + params->Bpl));
	else
		return 0;
}

/* Formula 1, R computation of ITU-T G.107 Specs
 */
double R(struct parameters *params) {
	if (!params)
		return 0;
	return Ro(params) - Is(params) - Id(params) - Ieef(params) + params->A;
}

void mos_init(struct parameters *params) {
	if (!params)
		return;
	params->SLR = 8;
	params->RLR = 2;
	params->STMRs = 15;
	params->Ds = 3;
	params->STMR = 15;
	params->Dr = 3;
	params->LSTRr = 18;
	params->TELR = 65;
	params->T = 0;
	params->WEPL = 110;
	params->Tr = 0;
	params->Ta = 0;
	params->Ie = 0;
	params->Bpl = 1;
	params->Ppl = 0;
	params->BurstR = 1;
	params->A = 0;
	params->Nc = -70;
	params->Ps = 35;
	params->Pr = 35;
	params->qdu = 1;
	params->Nfor = -64;
}

int main(int argc, char **argv) {
	int sel_opt, n_elements, ok;
	int fd_read1, fd_write; // File Handlers

	int is_verbose = 0, has_input = 0, has_output = 0,
		has_force = 0; // Boolean switches
	char *out_file; // Output file
	char line[MAX_LINE], ts_print[TSSIZE];

	int i, tmp_seq, n_seq;
	long long int current;
	struct timeval time1;
	struct parameters mos;
	double plp;
#ifdef DEBUG
	struct timeval t_start, t_end, t_res;
	long int t_tmp;
#endif
	p_flist_head hList;
	p_flow myFlow;

	// Inicialitzations
	fd_read1  = STDIN_FILENO;
	fd_write  = STDOUT_FILENO;
	bzero (line, MAX_LINE);

	// Main option parsing loop
	while(1) {
		sel_opt = getopt(argc, argv, "i:o:vfh");
		if (sel_opt == -1)
			break;
		switch (sel_opt) {
		case 'i':
			if (has_input) {
				fprintf(stderr,"Ingoring more input files\n");
				break;
			}
			fd_read1 = open(optarg,O_RDONLY);
			if (fd_read1 < 0) {
				perror("Error opening input file 1");
				exit(-1);
			}
			has_input = 1;
		break;
		case 'o':
			if (has_output)
				break;
			out_file = malloc(strlen(optarg+1));
			strncpy(out_file,optarg,strlen(optarg));
			has_output = 1;
		break;
		case 'v':
			is_verbose = 1;
		break;
		case 'f':
			has_force = 1;
		break;
		case 'h':
			usage();
			if (fd_read1 != STDIN_FILENO)
				close(fd_read1);
			exit(0);
		break;
		default:
			usage();
			exit(-2);
		break;
		} // End Switch sel_opt
	} // End While (option parsing)
	if (has_output) {
		if (!has_force) {
			fd_write = open(out_file, O_RDONLY);
			if (fd_write > 0) {
				close(fd_write);
				close(fd_read1);
				fprintf(stderr,"Error: Output file %s exists, use -f to override\n", out_file);
				exit(-1);
			}
		}
		fd_write = open(out_file, O_RDWR | O_CREAT, 0666);
		if (fd_write < 0) {
			perror("Error creating output file");
			exit(-1);
		}
	} // End if has_output

	i = 0;
#ifdef DEBUG
	gettimeofday(&t_start, NULL);
#endif
	hList = make_flow_list(fd_read1, is_verbose);
#ifdef DEBUG
	gettimeofday(&t_end, NULL);
	time_diff(&t_end, &t_start, &t_res);
	fprintf(stderr, "Flow list creation time: %d sec, %d usec\n", t_res.tv_sec, t_res.tv_usec);
	if (hList) {
		t_tmp = (t_res.tv_sec * 1000000l + t_res.tv_usec) / hList->elements;
		fprintf(stderr, "Total list elements: %d, %d sec and %d usec of computing time per packet\n",
			hList->elements, t_tmp / 1000000l, t_tmp % 1000000l);
		fprintf(stderr, "Maximum sequence number: %d\n", hList->max_seq);
	}
#endif

printf("-----------------------------------\nVigila falta l'efecte dels plugins\n-----------------------------------\n");

	ts2char(&hList->mean_delay, ts_print);
	plp = (double)hList->packet_losses;
	plp = plp / hList->test_packets;
	printf("Losses: %d, Packets: %d, PLP: %f, OOO Packets: %d, Mean Delay: %s, ",
	        hList->packet_losses, hList->test_packets, plp, hList->outoforder_packets, ts_print);
	ts2char(&hList->max_delay, ts_print);
	printf("Max Delay: %s, ", ts_print);
	ts2char(&hList->min_delay, ts_print);
	printf("Min Delay: %s\n", ts_print);

	mos_init(&mos);
	mos.Ppl = plp;

	ts2double(&hList->mean_delay, &(mos.T));

	printf("R_MOS = %f\n", R(&mos));

	close(fd_read1);
	close(fd_write);
	del_list(hList);
	return 0;
}

/* Procedure for displaying usage information */
int usage() {
	printf("Usage: ./mos -i <input_file> -o <output_file> [-f] [-v]\n");
	printf("	-i   input_file, if is omitted stdin will be used\n");
	printf("	-o   output_file, if is omitted stdout will be used\n");
	printf("	-f   for overwriting destination file\n");
	printf("	-v   Be verbose\n");
}
